from hive.core import HivePubSub
import rospy


class HiveRos(HivePubSub):

    def __init__(self, name):
        rospy.init_node(name, anonymous=True)
        self.publishers = {}

    def subscribe(self, topic, cls, callback):
        rospy.Subscriber(topic, cls, callback)

    def publish(self, topic, obj):
        publisher = self.publishers.get(topic, None)
        if publisher is None:
            publisher = rospy.Publisher(topic, type(obj), queue_size=1)
            self.publishers[topic] = publisher
        publisher.publish(obj)

    def spin(self):
        rospy.spin()

    def spin_once(self):
        pass



import time
import uuid

import stomp

from hive.core import HiveBytePubSub

class Listener(stomp.ConnectionListener):

    def __init__(self):
        self.subscriptions = {}

    def subscribe(self, topic, callback):
        subscriptions = self.subscriptions.get(topic, [])
        subscriptions.append(callback)
        self.subscriptions[topic] = subscriptions

    def on_message(self, headers, body):
        for cb in self.subscriptions[headers['destination']]:
            cb(body.encode("utf-8"))



class HiveStomp(HiveBytePubSub):

    def __init__(self):
        self.client = stomp.Connection()
        self.listener = Listener()
        self.client.set_listener('', self.listener)
        self.client.connect(wait=True)

    def subscribe_raw(self, topic, callback):
        self.client.subscribe(destination=topic, id=str(uuid.uuid4()), ack='auto')
        self.listener.subscribe(topic, callback)

    def publish_raw(self, topic, message):
        self.client.send(destination=topic, body=message)

    def spin(self):
        while True:
            time.sleep(0.01)

    def __del__(self):
        self.client.disconnect()




import jsonpickle
import time
import importlib


class HivePubSub:
    def subscribe(self, topic, cls, callback):
        '''
        Subscribe to a topic, with a callback expecting instances of type cls
        :param topic:
        :param cls:
        :param callback:
        :return:
        '''
        pass

    def publish(self, topic, obj):
        '''
        Subscribe to a topic, with a callback expecting instances of type cls
        :param topic:
        :param cls:
        :param callback:
        :return:
        '''
        pass

    def spin(self):
        '''
        Hand control to the pubsub and process incoming messages
        :return:
        '''
        while True:
            self.spin_once()

    def spin_once(self):
        '''
        Process the next message
        :return:
        '''
        time.sleep(0.001)


def json_encoder(cls, obj):
    return jsonpickle.encode(obj, unpicklable=False).encode("utf-8")


def json_decoder(cls, msg_bytes):
    d = jsonpickle.decode(msg_bytes.decode("utf-8"))
    if type(d) == cls:
        return d

    obj = cls()
    for k, v in d.items():
        setattr(obj, k, v)
    return obj


class HiveBytePubSub(HivePubSub):
    '''
    A specific Hive PubSub where all topics take bytes objects, and in-process
    encoders are used to convert the objects to and from byte streams.

    subscribe and publish call the encoders and then the subscribe_raw and publish_raw methods
    respectively. the *_raw methods always send and recieve bytes, and should be implemented
    in the subclasses

    Mqtt and Redis both use this as their base
    '''

    def __init__(self,
                 encoder=json_encoder,
                 decoder=json_decoder
                 ):
        self.encoder = encoder
        self.decoder = decoder

    def subscribe(self, topic, cls, callback):
        self.subscribe_raw(
            topic,
            lambda obj: callback(self.decoder(cls, obj))
        )

    def publish(self, topic, obj):
        self.publish_raw(topic, self.encoder(type(obj), obj))

    def subscribe_raw(self, topic, callback):
        """
        Subscribe to the bytes received on a topic

        :param topic: topic name
        :param callback: callback function to execute - cb(bytes)
        :return: None
        """
        pass

    def publish_raw(self, topic, message):
        """
        Publish raw bytes to a topic
        :param topic: topic to subscribe
        :param message: message bytes to publish
        :return:
        """
        pass


class Publisher:
    '''
    Convenience class to collect a pubsub instance and a topic for easy repeat publishing
    '''

    def __init__(
            self,
            pubsub,
            topic,
    ):
        self.pubsub = pubsub
        self.topic = topic

    def publish(self, obj):
        self.pubsub.publish(self.topic, obj)


class HiveWorker:
    pubsub_cls = HivePubSub

    def __init__(self, *args, **kwargs):
        self.pubsub = HiveWorker.pubsub_cls(*args, **kwargs)

    def spin(self):
        try:
            self.pubsub.spin()
        except KeyboardInterrupt:
            pass

    def publisher(self, topic):
        return Publisher(self.pubsub, topic)

    def subscribe(self, topic, cls, callback):
        return self.pubsub.subscribe(topic, cls, callback)

    @staticmethod
    def load_from_config(filename="hive.json"):
        with open(filename) as f:
            json = f.read()
        options = jsonpickle.decode(json)

        module_name, _, class_name = options["pubsub"].rpartition('.')
        module = importlib.import_module(module_name)
        cls = getattr(module, class_name)
        HiveWorker.pubsub_cls = cls
        return HiveWorker(**options["init"])



import logging

import coilmq.start

logger = logging.getLogger(__name__)

try:
    coilmq.start.main()
except (KeyboardInterrupt, SystemExit):
    pass
except Exception as e:
    logger.error("Server terminated due to error: %s" % e)
    logger.exception(e)



from hive.core import HiveBytePubSub
import redis


class HiveRedis(HiveBytePubSub):

    def __init__(self, host="localhost", port=6379, db=0):
        self.client = redis.Redis(host=host, port=port, db=db)
        self.sub = self.client.pubsub()
        self.subscriptions = {}

    def subscribe_raw(self, topic, callback):
        self.sub.subscribe(topic)
        subscriptions = self.subscriptions.get(topic, [])
        subscriptions.append(callback)
        self.subscriptions[topic] = subscriptions

    def publish_raw(self, topic, message):
        self.client.publish(topic, message)

    def spin(self):
        while True:
            msg = self.sub.get_message()
            if msg and msg["type"] == "message":
                for cb in self.subscriptions[msg["channel"].decode("utf-8")]:
                    cb(msg["data"])




import paho.mqtt.client as mqtt
from hive.core import HiveBytePubSub, json_encoder, json_decoder


class HiveMqtt(HiveBytePubSub):

    def __init__(self,
                 # Connection options
                 host="localhost",
                 port=1883,
                 keepalive=60,
                 bind_address="",
                 bind_port=0,
                 clean_start=mqtt.MQTT_CLEAN_START_FIRST_ONLY,
                 properties=None,

                 # client options
                 client_id="",
                 clean_session=None,
                 userdata=None,
                 protocol=mqtt.MQTTv311,
                 transport="tcp",

                 # encoder options
                 encoder=json_encoder,
                 decoder=json_decoder
                 ):
        super().__init__(encoder, decoder)
        self.client = mqtt.Client(client_id, clean_session, userdata, protocol, transport)
        self.client.connect(host, port, keepalive, bind_address, bind_port, clean_start, properties)
        self.client.on_message = self.on_message()  # we use this to close over self so we can access the subscription list
        self.subscriptions = {}

    def on_message(self):
        def client_on_message(client, userdata, message):
            for cb in self.subscriptions[message.topic]:
                cb(message.payload)

        return client_on_message

    def subscribe_raw(self, topic, callback):
        self.client.subscribe(topic)
        subscriptions = self.subscriptions.get(topic, [])
        subscriptions.append(callback)
        self.subscriptions[topic] = subscriptions

    def publish_raw(self, topic, message):
        self.client.publish(topic, message)

    def spin(self):
        self.client.loop_forever()



import msgpack

packer = msgpack.Packer(default=lambda o: o.__dict__)


def msgpack_encoder(cls, obj):
    return packer.pack(obj)


def msgpack_decoder(cls, msg_bytes):
    d = msgpack.unpackb(msg_bytes)
    if type(d) == cls:
        return d

    obj = cls()
    for k, v in d.items():
        setattr(obj, k, v)
    return obj



import setuptools


with open('README.md') as f:
    README = f.read()

setuptools.setup(
    author="James McMahon",
    author_email="james1345@googlemail.com",
    name='hive-py',
    license="MIT",
    description='A library to make multi-process and multi-machine programs easy to write',
    version='v0.0.6',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://bitbucket.org/james1345/hive/',
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    install_requires=['paho-mqtt', 'jsonpickle'],
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
    ],
)